(* General module to embed content.
   Usage: 'ocaml shell/embed.ml > opamEmbedModule.ml'
*)
let () =
  let setup = Sys.argv.(1) in
  let content_setup =
    let ic = open_in_bin setup in
    let len = in_channel_length ic in
    let content = really_input_string ic len in
    close_in ic;
    content
  in
  Printf.printf
    {|(* THIS FILE IS AUTOMATICALLY GENERATED, EDIT configure.ac & shell/embed.ml INSTEAD *)
let content = Some "%s"
|}
    (String.escaped content_setup)
