(**************************************************************************)
(*                                                                        *)
(*    Copyright 2026 OCamlPro                                             *)
(*                                                                        *)
(*  All rights reserved. This file is distributed under the terms of the  *)
(*  GNU Lesser General Public License version 2.1, with the special       *)
(*  exception on linking described in the file LICENSE.                   *)
(*                                                                        *)
(**************************************************************************)

let () =
  OpamUnit.run_tests @@ fun ctxt ->
  TestOpamFilename.test ~ctxt ();
  TestOpamSystem.test ~ctxt ();
  ()
